-- bayer4_bm16.lua : converts an image into BM16
-- mode for thomson machines (MO6,TO8,TO9,TO9+)
-- using bayer matrix and a special palette.
--
-- Version 1.1 Copyright 2016 Samuel Devulder
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; version 2
-- of the License. See <http://www.gnu.org/licenses/>

-- This is my first code in lua, so excuse any bad coding practices.

-- use gamma-correction ("true"" recommended, but sometimes 
-- "false" gives smoother results YMMV.)
local with_gamma_corr = with_gamma_corr or true

-- use a zig zag. If false (recommended value), this gives 
-- a raster look and feel
local with_zig_zag = with_zig_zag or false

-- debug: displays histograms
local debug = false

-- enhance luminosity since our mode divide it by two
local enhance_lum = enhance_lum or true

-- use fixed levels (default=false, give better result)
local fixed_levels = fixed_levels or false

-- get screen size
local screen_w, screen_h = getpicturesize()

-- if screen isn't big enough, then increase its size
if screen_w<320 and screen_h<200 then 
	setpicturesize(320,200); 
	finalizepicture(); 
end

run("thomson.lua")
run("pixel.lua")

-- Converts thomson coordinates (0-159,0-99) into screen coordinates
local function thom2screen(x,y)
	local i,j;
	if screen_w/screen_h < 1.6 then
		i = x*screen_h/100
		j = y*screen_h/100
	else
		i = x*screen_w/160
		j = y*screen_w/160
	end
	return math.floor(i), math.floor(j)
end

-- return the pixel @(x,y) in linear space corresonding to the thomson screen (x in 0-159, y in 0-99)
local function getLinearPixel(x,y)
	local x1,y1 = thom2screen(x,y)
	local x2,y2 = thom2screen(x+1,y+1)
	if x2==x1 then x2=x1+1 end
	if y2==y1 then y2=y1+1 end

	local p,i,j = Pixel:new(0,0,0);
	for i=x1,x2-1 do
		for j=y1,y2-1 do
			p:add(getLinearPicturePixel(i,j))
		end
	end

	return p:div((y2-y1)*(x2-x1)):floor()
end

--[[ make a bayer matrix
function bayer(matrix)
	local m,n=#matrix,#matrix[1]
	local r,i,j = {}
	for j=1,m*2 do
		local t = {}
		for i=1,n*2 do t[i]=0; end
		r[j] = t;
	end
	
	-- 0 3
	-- 2 1
	for j=1,m do
		for i=1,n do
			local v = 4*matrix[j][i]
			r[m*0+j][n*0+i] = v-3
			r[m*1+j][n*1+i] = v-2
			r[m*1+j][n*0+i] = v-1
			r[m*0+j][n*1+i] = v-0
		end
	end
	
	return r;
end
--]]

-- dither matrix
local dither = {
	-- vac8: kind of like FS result
	-- {35,57,19,55,7,51,4,21},
	-- {29,6,41,27,37,17,59,45},
	-- {61,15,53,12,62,25,33,9},
	-- {23,39,31,49,2,47,13,43},
	-- {3,52,8,22,36,58,20,56},
	-- {38,18,60,46,30,5,42,28},
	-- {63,26,34,11,64,16,54,10},
	-- {14,48,1,44,24,40,32,50}
	
	-- bayer 4x4
	{1,9,3,11},
	{13,5,15,7},
	{4,12,2,10},
	{16,8,14,6}
};

-- put dither matrix in 0..1 range
do
	local dx,dy = #dither,#dither[1]
	local norm  = 1/(1+dx*dy)
	for i=1,dx do
		for j=1,dy do
			dither[i][j] = dither[i][j]*norm
		end
	end
end

-- get color statistics
local stat = {};
function stat:clear() 
	self.r = {}
	self.g = {}
	self.b = {}
	for i=1,16 do self.r[i] = 0; self.g[i] = 0; self.b[i] = 0; end
end
function stat:update(px) 
	local pc2to = thomson.levels.pc2to
	local r,g,b=pc2to[px.r], pc2to[px.g], pc2to[px.b];
	self.r[r] = self.r[r] + 1;
	self.g[g] = self.g[g] + 1;
	self.b[b] = self.b[b] + 1;
end
stat:clear();
for y = 0,99 do
	for x = 0,159 do
		stat:update(getLinearPixel(x,y))
	end
end

-- find key color
local function findLvls(name, stat, levels)
	local tot, max = 0,0;
	for _,t in ipairs(stat) do
		max = math.max(t,max); 
		tot = tot + t;
	end	
	local function f(tot,levels,res)
		local acc=0
		for _,t in ipairs(levels) do
			for i=(res[res.n] or 0)+1,16 do
				acc = acc + stat[i]
				if acc>=t*tot then
					tot = tot-acc; acc = 0;
					res.n = res.n+1
					res[res.n] = i;
					break
				end
			end
		end
		return res
	end
	local res=f(tot, levels,{n=0})

	-- not enough elements found, use alter algorithm
	if res.n~=#levels then
		local deb,fin = res[1],res[res.n]
		local all = tot
		for i=1,deb  do all=all-stat[i] end
		for i=fin,16 do all=all-stat[i] end
		local levs = {}
		for i=2,#levels-1 do levs[i-1]=levels[i]; end
		res = f(all, levs, {deb, n=1})
		res.n = res.n+1
		res[res.n] = fin
	end
	
	-- sanity
	while res.n<#levels do
		res.n = res.n + 1
		res[res.n] = res[res.n-1]
	end
	
	-- convert to full array
	res.n=nil
	
	-- debug
	if debug then
		local txt = ""
		for _,i in ipairs(res) do
			txt = txt .. i .. " "
		end
		for i,t in ipairs(stat) do
			txt = txt .. "\n" .. string.format("%s%2d:%3d%% ", name, i, math.floor(100*t/tot))
			for _=0,math.floor(26*t/tot)-1 do txt=txt .. "X"; end
		end
		messagebox(txt)
	end
	
	return res
end

local grnPerc,redPerc,bluPerc
if with_gamma_corr then
	-- grnPerc = {0,0.20,0.20,0.30,0.80}
	-- redPerc = {0,0.20,0.30,0.80}
	-- bluPerc = {0,0.30,0.80}
	grnPerc = {0,0.30,0.30,0.50,0.90}
	redPerc = {0,0.30,0.50,0.90}
	bluPerc = {0,0.45,0.90}
else
	grnPerc = {0,0.20,0.20,0.50,0.85}
	redPerc = {0,0.20,0.50,0.85}
	bluPerc = {0,0.50,0.85}
end

grnLvls = findLvls("GRN",stat.g,grnPerc)
redLvls = findLvls("RED",stat.r,redPerc)
bluLvls = findLvls("BLU",stat.b,bluPerc)

if fixed_levels then
	findLvls = function(name, stat, levels)
		local res,max = {},#levels-1
		for i=0,max do
			local p = Pixel:new(255*i/max,0,0):toPC():floor().r
			res[i+1] = thomson.levels.pc2to[p]
		end
		return res
	end
end

-- enhance luminosity since our mode divide it by two
local gain = 1
if enhance_lum then
	local maxLvl = 0
	for _,i in ipairs(redLvls) do maxLvl = math.max(maxLvl,i); end
	for _,i in ipairs(grnLvls) do maxLvl = math.max(maxLvl,i); end
	for _,i in ipairs(bluLvls) do maxLvl = math.max(maxLvl,i); end
	gain = math.min(2,255/thomson.levels.linear[maxLvl])

	-- redo stat with enhanced levels
	stat:clear();
	for y = 0,99 do
		for x = 0,159 do
			stat:update(getLinearPixel(x,y):mul(gain):floor())
		end
	end
	
	grnLvls = findLvls("GRN",stat.g,grnPerc)
	redLvls = findLvls("RED",stat.r,redPerc)
	bluLvls = findLvls("BLU",stat.b,bluPerc)
end

local function pset(x,y,px)
	local thr = dither[1+(y % #dither)][1+(x % #dither[1])]
	local function dither(val,thr,lvls)
		local i=#lvls
		local a,b = thomson.levels.linear[lvls[i]],1e30
		while i>1 and val<a do 
			i=i-1;
			a,b=thomson.levels.linear[lvls[i]],a; 
		end
		return i + ((val-a)>=thr*(b-a) and 0 or -1)
	end

	local r = dither(px.r, thr, redLvls);
	local g = dither(px.g, thr, grnLvls);
	local b = dither(px.b, thr, bluLvls);
	
	local i = r + b*4
	local j = g==0 and 0 or (11 + g)
	
	if with_zig_zag and x%2==1 then
		thomson.pset(x,y*2+0,j)
		thomson.pset(x,y*2+1,i)
	else
		thomson.pset(x,y*2+0,i)
		thomson.pset(x,y*2+1,j)
	end
end

-- BM16 mode
thomson.setBM16()

-- define palette
for i=0,15 do
	local r,g,b=0,0,0
	if i<12 then
		-- r = bit32.band(i,3)
		-- b = bit32.rshift(i,2)
		b = math.floor(i/4)
		r = i-4*b
	else
		g = i-11
	end
	r,g,b=redLvls[r+1],grnLvls[g+1],bluLvls[b+1]
	thomson.palette(i,b*256+g*16+r-273)
end

-- convert picture
for y = 0,99 do
	for x = 0,159 do
		pset(x,y, getLinearPixel(x,y):mul(gain):floor())
	end
end

-- refresh screen
setpicturesize(320,200)
thomson.updatescreen()
finalizepicture()

-- save picture
do
	local function exist(file)
		local f=io.open(file,'rb')
		if not f then return false else io.close(f); return true; end
	end
	local name,path = getfilename()
	local mapname = string.gsub(name,"%.%w*$","") .. ".map"
	local fullname = path .. '/' .. mapname
	-- fullname = 'D:/tmp/toto.map'
	local ok = not exist(fullname)
	if not ok then
		selectbox("Ovr " .. mapname .. "?", "Yes", function() ok = true; end, "No", function() ok = false; end)
	end
	if ok then thomson.savep(fullname) end
end
